#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import copy
from unittest import mock

import pytest
from botocore.exceptions import ClientError

from airflow.exceptions import AirflowException, TaskDeferred
from airflow.providers.amazon.aws.hooks.sagemaker import SageMakerHook
from airflow.providers.amazon.aws.links.sagemaker import SageMakerTransformJobLink
from airflow.providers.amazon.aws.operators import sagemaker
from airflow.providers.amazon.aws.operators.sagemaker import SageMakerTransformOperator
from airflow.providers.amazon.aws.triggers.sagemaker import SageMakerTrigger
from airflow.providers.common.compat.openlineage.facet import Dataset
from airflow.providers.openlineage.extractors import OperatorLineage

from unit.amazon.aws.utils.test_template_fields import validate_template_fields

EXPECTED_INTEGER_FIELDS: list[list[str]] = [
    ["Transform", "TransformResources", "InstanceCount"],
    ["Transform", "MaxConcurrentTransforms"],
    ["Transform", "MaxPayloadInMB"],
]

CREATE_TRANSFORM_PARAMS: dict = {
    "TransformJobName": "job_name",
    "ModelName": "model_name",
    "MaxConcurrentTransforms": "12",
    "MaxPayloadInMB": "6",
    "BatchStrategy": "MultiRecord",
    "TransformInput": {"DataSource": {"S3DataSource": {"S3DataType": "S3Prefix", "S3Uri": "s3_uri"}}},
    "TransformOutput": {"S3OutputPath": "output_path"},
    "TransformResources": {"InstanceType": "ml.m4.xlarge", "InstanceCount": "3"},
}
CREATE_TRANSFORM_PARAMS_INTEGER_FIELDS: dict = {
    "TransformJobName": "job_name",
    "ModelName": "model_name",
    "MaxConcurrentTransforms": 12,
    "MaxPayloadInMB": 6,
    "BatchStrategy": "MultiRecord",
    "TransformInput": {"DataSource": {"S3DataSource": {"S3DataType": "S3Prefix", "S3Uri": "s3_uri"}}},
    "TransformOutput": {"S3OutputPath": "output_path"},
    "TransformResources": {"InstanceType": "ml.m4.xlarge", "InstanceCount": 3},
}

CREATE_MODEL_PARAMS: dict = {
    "ModelName": "model_name",
    "PrimaryContainer": {"Image": "test_image", "ModelDataUrl": "output_path"},
    "ExecutionRoleArn": "arn:aws:iam:role/test-role",
}

CONFIG: dict = {"Model": CREATE_MODEL_PARAMS, "Transform": CREATE_TRANSFORM_PARAMS}

MOCK_UNIX_TIME: int = 1234567890123456789  # reproducible time for testing time.time_ns()


class TestSageMakerTransformOperator:
    def setup_method(self):
        self.sagemaker = SageMakerTransformOperator(
            task_id="test_sagemaker_operator",
            config=copy.deepcopy(CONFIG),
            wait_for_completion=False,
            check_interval=5,
            check_if_model_exists=False,
        )

    @mock.patch.object(SageMakerHook, "describe_transform_job")
    @mock.patch.object(SageMakerHook, "create_model")
    @mock.patch.object(SageMakerHook, "describe_model")
    @mock.patch.object(SageMakerHook, "create_transform_job")
    @mock.patch.object(sagemaker, "serialize", return_value="")
    def test_integer_fields(self, _, mock_create_transform, __, ___, mock_desc):
        mock_desc.side_effect = [
            ClientError({"Error": {"Code": "ValidationException"}}, "op"),
            {"ModelName": "model_name"},
        ]
        mock_create_transform.return_value = {
            "TransformJobArn": "test_arn",
            "ResponseMetadata": {"HTTPStatusCode": 200},
        }
        self.sagemaker.execute(None)
        assert self.sagemaker.integer_fields == EXPECTED_INTEGER_FIELDS
        for key1, key2, *key3 in EXPECTED_INTEGER_FIELDS:
            if key3:
                (key3,) = key3
                assert self.sagemaker.config[key1][key2][key3] == int(self.sagemaker.config[key1][key2][key3])
            else:
                self.sagemaker.config[key1][key2] == int(self.sagemaker.config[key1][key2])

    @mock.patch.object(SageMakerHook, "describe_transform_job")
    @mock.patch.object(SageMakerHook, "create_model")
    @mock.patch.object(SageMakerHook, "describe_model")
    @mock.patch.object(SageMakerHook, "create_transform_job")
    @mock.patch.object(sagemaker, "serialize", return_value="")
    def test_execute(self, _, mock_transform, __, mock_model, mock_desc):
        mock_desc.side_effect = [
            ClientError({"Error": {"Code": "ValidationException"}}, "op"),
            {"ModelName": "model_name"},
        ]
        mock_transform.return_value = {
            "TransformJobArn": "test_arn",
            "ResponseMetadata": {"HTTPStatusCode": 200},
        }
        self.sagemaker.execute(None)
        mock_model.assert_called_once_with(CREATE_MODEL_PARAMS)
        mock_transform.assert_called_once_with(
            CREATE_TRANSFORM_PARAMS_INTEGER_FIELDS,
            wait_for_completion=False,
            check_interval=5,
            max_ingestion_time=None,
        )

    @mock.patch.object(SageMakerHook, "describe_transform_job")
    @mock.patch.object(SageMakerHook, "create_model")
    @mock.patch.object(SageMakerHook, "describe_model")
    @mock.patch.object(SageMakerHook, "create_transform_job")
    # @mock.patch.object(sagemaker, "serialize", return_value="")
    def test_log_correct_url(self, mock_transform, __, ___, mock_desc):
        region = "us-east-1"
        job_name = CONFIG["Transform"]["TransformJobName"]
        mock_desc.side_effect = [
            ClientError({"Error": {"Code": "ValidationException"}}, "op"),
            {"ModelName": "model_name"},
        ]
        mock_transform.return_value = {
            "TransformJobArn": "test_arn",
            "ResponseMetadata": {"HTTPStatusCode": 200},
        }

        aws_domain = SageMakerTransformJobLink.get_aws_domain("aws")
        job_run_url = (
            f"https://console.{aws_domain}/sagemaker/home?region={region}#/transform-jobs/{job_name}"
        )

        with mock.patch.object(self.sagemaker.log, "info") as mock_log_info:
            self.sagemaker.execute(None)
            # assert job_run_id == JOB_RUN_ID
        mock_log_info.assert_any_call("You can monitor this SageMaker Transform job at %s", job_run_url)

    @mock.patch.object(SageMakerHook, "describe_transform_job")
    @mock.patch.object(SageMakerHook, "create_model")
    @mock.patch.object(SageMakerHook, "create_transform_job")
    def test_execute_with_failure(self, mock_transform, _, mock_desc):
        mock_desc.side_effect = [
            ClientError({"Error": {"Code": "ValidationException"}}, "op"),
            None,
        ]
        mock_transform.return_value = {
            "TransformJobArn": "test_arn",
            "ResponseMetadata": {"HTTPStatusCode": 404},
        }
        with pytest.raises(AirflowException):
            self.sagemaker.execute(None)

    @mock.patch.object(SageMakerHook, "describe_transform_job")
    @mock.patch.object(SageMakerHook, "create_transform_job")
    @mock.patch.object(SageMakerHook, "create_model")
    @mock.patch.object(SageMakerHook, "describe_model")
    @mock.patch.object(sagemaker, "serialize", return_value="")
    def test_execute_with_check_if_job_exists(self, _, __, ___, mock_transform, mock_desc):
        mock_desc.side_effect = [
            ClientError({"Error": {"Code": "ValidationException"}}, "op"),
            {"ModelName": "model_name"},
        ]
        mock_transform.return_value = {
            "TransformJobArn": "test_arn",
            "ResponseMetadata": {"HTTPStatusCode": 200},
        }
        self.sagemaker.execute(None)
        mock_transform.assert_called_once_with(
            CREATE_TRANSFORM_PARAMS_INTEGER_FIELDS,
            wait_for_completion=False,
            check_interval=5,
            max_ingestion_time=None,
        )

    @mock.patch.object(SageMakerHook, "describe_transform_job")
    @mock.patch.object(SageMakerHook, "create_transform_job")
    @mock.patch.object(SageMakerHook, "create_model")
    @mock.patch.object(SageMakerHook, "describe_model")
    @mock.patch.object(sagemaker, "serialize", return_value="")
    def test_execute_without_check_if_job_exists(self, _, __, ___, mock_transform, ____):
        mock_transform.return_value = {
            "TransformJobArn": "test_arn",
            "ResponseMetadata": {"HTTPStatusCode": 200},
        }
        self.sagemaker.check_if_job_exists = False
        self.sagemaker.execute(None)
        mock_transform.assert_called_once_with(
            CREATE_TRANSFORM_PARAMS_INTEGER_FIELDS,
            wait_for_completion=False,
            check_interval=5,
            max_ingestion_time=None,
        )

    @mock.patch(  # since it is divided by 1000000000, the added timestamp should be 1234567890.
        "airflow.providers.amazon.aws.operators.sagemaker.time.time_ns", return_value=MOCK_UNIX_TIME
    )
    @mock.patch.object(
        SageMakerHook, "describe_transform_job", return_value={"ModelName": "model_name-1234567890"}
    )
    @mock.patch.object(
        SageMakerHook,
        "create_transform_job",
        return_value={
            "ResponseMetadata": {"HTTPStatusCode": 200},
        },
    )
    @mock.patch.object(SageMakerHook, "create_model")
    @mock.patch.object(
        SageMakerHook,
        "describe_model",
        side_effect=[
            None,
            ClientError({"Error": {"Code": "ValidationException"}}, "op"),
            "model_name-1234567890",
        ],
    )
    @mock.patch.object(sagemaker, "serialize", return_value="")
    def test_when_model_already_exists_it_should_add_timestamp_to_model_name(
        self, _, mock_describe_model, mock_create_model, __, ___, timestamp_mock
    ):
        self.sagemaker.check_if_job_exists = False
        self.sagemaker.check_if_model_exists = True
        model_config = {"ModelName": "model_name"}
        self.sagemaker.config["Model"] = model_config

        self.sagemaker.execute(None)

        mock_describe_model.assert_has_calls(
            [mock.call("model_name"), mock.call("model_name-1234567890"), mock.call("model_name-1234567890")]
        )
        mock_create_model.assert_called_once_with({"ModelName": "model_name-1234567890"})

    @mock.patch.object(SageMakerHook, "describe_transform_job")
    @mock.patch.object(SageMakerHook, "create_transform_job")
    @mock.patch.object(SageMakerHook, "create_model")
    @mock.patch.object(SageMakerHook, "describe_model")
    @mock.patch.object(sagemaker, "serialize", return_value="")
    def test_when_model_already_exists_it_should_raise_airflow_exception(
        self, _, mock_describe_model, mock_create_model, __, ___
    ):
        mock_describe_model.side_effect = [None]
        self.sagemaker.check_if_job_exists = False
        self.sagemaker.check_if_model_exists = True
        self.sagemaker.action_if_model_exists = "fail"
        model_config = {"ModelName": "model_name"}
        self.sagemaker.config["Model"] = model_config

        with pytest.raises(AirflowException) as context:
            self.sagemaker.execute(None)

        assert str(context.value) == "A SageMaker model with name model_name already exists."
        mock_describe_model.assert_called_once_with("model_name")
        mock_create_model.assert_not_called()

    @mock.patch.object(SageMakerHook, "describe_transform_job", return_value={"ModelName": "model_name"})
    @mock.patch.object(
        SageMakerHook,
        "create_transform_job",
        return_value={
            "ResponseMetadata": {"HTTPStatusCode": 200},
        },
    )
    @mock.patch.object(SageMakerHook, "create_model")
    @mock.patch.object(SageMakerHook, "describe_model")
    @mock.patch.object(sagemaker, "serialize", return_value="")
    def test_execute_without_check_if_model_exists(self, _, mock_describe_model, mock_create_model, __, ___):
        self.sagemaker.check_if_job_exists = False
        self.sagemaker.check_if_model_exists = False
        model_config = {"ModelName": "model_name"}
        self.sagemaker.config["Model"] = model_config
        self.sagemaker._get_unique_model_name = mock.Mock()

        self.sagemaker.execute(None)

        mock_create_model.assert_called_once_with(model_config)
        mock_describe_model.assert_called_once_with("model_name")
        self.sagemaker._get_unique_model_name.assert_not_called()

    @mock.patch.object(SageMakerTransformOperator, "_get_unique_name")
    def test_get_unique_model_name_calls_get_unique_name_correctly(self, get_unique_name_mock):
        def describe_func():
            pass

        self.sagemaker._get_unique_model_name("model_name", True, describe_func)

        get_unique_name_mock.assert_called_once_with(
            "model_name",
            True,
            describe_func,
            self.sagemaker._check_if_model_exists,
            "model",
        )

    @mock.patch.object(SageMakerTransformOperator, "_check_if_resource_exists")
    def test_check_if_model_exists_calls_check_if_resource_exists_correctly(self, check_resource_exists_mock):
        def describe_func():
            pass

        self.sagemaker._check_if_model_exists("model_name", describe_func)

        check_resource_exists_mock.assert_called_once_with("model_name", "model", describe_func)

    @mock.patch("airflow.providers.amazon.aws.operators.sagemaker.SageMakerTransformOperator.defer")
    @mock.patch.object(
        SageMakerHook,
        "describe_transform_job",
        return_value={"TransformJobStatus": "Failed", "FailureReason": "it failed"},
    )
    @mock.patch.object(SageMakerHook, "create_transform_job")
    @mock.patch.object(SageMakerHook, "create_model")
    def test_operator_failed_before_defer(self, _, mock_transform, mock_describe_transform_job, mock_defer):
        mock_transform.return_value = {
            "TransformJobArn": "test_arn",
            "ResponseMetadata": {"HTTPStatusCode": 200},
        }
        self.sagemaker.deferrable = True
        self.sagemaker.wait_for_completion = True
        self.sagemaker.check_if_job_exists = False

        with pytest.raises(AirflowException):
            self.sagemaker.execute(context=None)
        assert not mock_defer.called

    @mock.patch("airflow.providers.amazon.aws.operators.sagemaker.SageMakerTransformOperator.defer")
    @mock.patch.object(SageMakerHook, "describe_model")
    @mock.patch.object(
        SageMakerHook, "describe_transform_job", return_value={"TransformJobStatus": "Completed"}
    )
    @mock.patch.object(SageMakerHook, "create_transform_job")
    @mock.patch.object(SageMakerHook, "create_model")
    def test_operator_complete_before_defer(
        self, _, mock_transform, mock_describe_transform_job, mock_describe_model, mock_defer
    ):
        mock_transform.return_value = {
            "TransformJobArn": "test_arn",
            "ResponseMetadata": {"HTTPStatusCode": 200},
        }
        mock_describe_model.return_value = {"PrimaryContainer": {"ModelPackageName": "package-name"}}
        self.sagemaker.deferrable = True
        self.sagemaker.wait_for_completion = True
        self.sagemaker.check_if_job_exists = False

        self.sagemaker.execute(context=None)
        assert not mock_defer.called

    @mock.patch.object(
        SageMakerHook, "describe_transform_job", return_value={"TransformJobStatus": "InProgress"}
    )
    @mock.patch.object(SageMakerHook, "create_transform_job")
    @mock.patch.object(SageMakerHook, "create_model")
    def test_operator_defer(self, _, mock_transform, mock_describe_transform_job):
        mock_transform.return_value = {
            "TransformJobArn": "test_arn",
            "ResponseMetadata": {"HTTPStatusCode": 200},
        }
        self.sagemaker.deferrable = True
        self.sagemaker.wait_for_completion = True
        self.sagemaker.check_if_job_exists = False

        with pytest.raises(TaskDeferred) as exc:
            self.sagemaker.execute(context=None)

        assert isinstance(exc.value.trigger, SageMakerTrigger), "Trigger is not a SagemakerTrigger"

    @mock.patch.object(SageMakerHook, "describe_transform_job")
    @mock.patch.object(SageMakerHook, "create_model")
    @mock.patch.object(SageMakerHook, "describe_model")
    @mock.patch.object(SageMakerHook, "get_conn")
    @mock.patch.object(SageMakerHook, "create_transform_job")
    def test_operator_lineage_data(self, mock_transform, mock_conn, mock_model, _, mock_desc):
        self.sagemaker.check_if_job_exists = False
        mock_conn.return_value.describe_model_package.return_value = {
            "InferenceSpecification": {"Containers": [{"ModelDataUrl": "s3://model-bucket/model-path"}]},
        }
        mock_model.return_value = {"PrimaryContainer": {"ModelPackageName": "package-name"}}
        mock_desc.return_value = {
            "TransformInput": {"DataSource": {"S3DataSource": {"S3Uri": "s3://input-bucket/input-path"}}},
            "TransformOutput": {"S3OutputPath": "s3://output-bucket/output-path"},
            "ModelName": "model_name",
        }
        mock_transform.return_value = {
            "TransformJobArn": "test_arn",
            "ResponseMetadata": {"HTTPStatusCode": 200},
        }
        self.sagemaker.execute(None)
        assert self.sagemaker.get_openlineage_facets_on_complete(None) == OperatorLineage(
            inputs=[
                Dataset(namespace="s3://input-bucket", name="input-path"),
                Dataset(namespace="s3://model-bucket", name="model-path"),
            ],
            outputs=[Dataset(namespace="s3://output-bucket", name="output-path")],
        )

    def test_template_fields(self):
        validate_template_fields(self.sagemaker)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import pytest
from moto import mock_aws

from airflow.providers.amazon.aws.hooks.eventbridge import EventBridgeHook


@mock_aws
class TestEventBridgeHook:
    def test_conn_returns_a_boto3_connection(self):
        hook = EventBridgeHook(aws_conn_id="aws_default")
        assert hook.conn is not None

    def test_put_rule(self):
        hook = EventBridgeHook(aws_conn_id="aws_default")
        response = hook.put_rule(
            name="test",
            event_pattern='{"source": ["aws.s3"]}',
            state="ENABLED",
        )
        assert "RuleArn" in response

    def test_put_rule_with_bad_json_fails(self):
        hook = EventBridgeHook(aws_conn_id="aws_default")
        with pytest.raises(ValueError):
            hook.put_rule(
                name="test",
                event_pattern="invalid json",
                state="ENABLED",
            )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

import inspect
from importlib import import_module
from pathlib import Path

import pytest

from airflow.exceptions import AirflowOptionalProviderFeatureException
from airflow.providers.amazon.aws.hooks.base_aws import AwsGenericHook

BASE_AWS_HOOKS = ["AwsGenericHook", "AwsBaseHook"]
ALLOWED_THICK_HOOKS_PARAMETERS: dict[str, set[str]] = {
    # This list should only be reduced not extended with new parameters,
    # unless there is an exceptional reason.
    "AthenaHook": {"sleep_time", "log_query"},
    "AthenaSQLHook": {"athena_conn_id"},
    "BatchClientHook": {"status_retries", "max_retries"},
    "BatchWaitersHook": {"waiter_config"},
    "DataSyncHook": {"wait_interval_seconds"},
    "DynamoDBHook": {"table_name", "table_keys"},
    "EC2Hook": {"api_type"},
    "ElastiCacheReplicationGroupHook": {
        "exponential_back_off_factor",
        "max_retries",
        "initial_poke_interval",
    },
    "EmrHook": {"emr_conn_id"},
    "EmrContainerHook": {"virtual_cluster_id"},
    "FirehoseHook": {"delivery_stream"},
    "GlueJobHook": {
        "job_name",
        "concurrent_run_limit",
        "job_poll_interval",
        "create_job_kwargs",
        "desc",
        "iam_role_arn",
        "s3_bucket",
        "iam_role_name",
        "update_config",
        "retry_limit",
        "num_of_dpus",
        "script_location",
    },
    "S3Hook": {"transfer_config_args", "aws_conn_id", "extra_args"},
}


def get_aws_hooks_modules():
    """Parse Amazon Provider metadata and find all hooks based on `AwsGenericHook` and return it."""
    import airflow.providers.amazon.aws.hooks as aws_hooks

    hooks_dir = Path(aws_hooks.__path__[0])
    if not hooks_dir.exists():
        msg = f"Amazon Provider hooks directory not found: {hooks_dir.__fspath__()!r}"
        raise FileNotFoundError(msg)
    elif not hooks_dir.is_dir():
        raise NotADirectoryError(hooks_dir.__fspath__())

    for module in sorted(hooks_dir.glob("*.py")):
        name = module.stem
        if name.startswith("_"):
            continue
        module_string = f"airflow.providers.amazon.aws.hooks.{name}"

        yield pytest.param(module_string, id=name)


def get_aws_hooks_from_module(hook_module: str) -> list[tuple[type[AwsGenericHook], str]]:
    try:
        imported_module = import_module(hook_module)
    except AirflowOptionalProviderFeatureException as ex:
        pytest.skip(str(ex))
    else:
        hooks = []
        for name, o in vars(imported_module).items():
            if name in BASE_AWS_HOOKS:
                continue

            if isinstance(o, type) and o.__module__ != "builtins" and issubclass(o, AwsGenericHook):
                hooks.append((o, name))
        return hooks


def validate_hook(hook: type[AwsGenericHook], hook_name: str, hook_module: str) -> tuple[bool, str | None]:
    hook_extra_parameters = set()
    for k, v in inspect.signature(hook.__init__).parameters.items():
        if v.kind == inspect.Parameter.VAR_POSITIONAL:
            k = "*args"
        elif v.kind == inspect.Parameter.VAR_KEYWORD:
            k = "**kwargs"

        hook_extra_parameters.add(k)
    hook_extra_parameters.difference_update({"self", "*args", "**kwargs"})

    allowed_parameters = ALLOWED_THICK_HOOKS_PARAMETERS.get(hook_name, set())
    if allowed_parameters:
        # Remove historically allowed parameters for Thick Wrapped Hooks
        hook_extra_parameters -= allowed_parameters

    if not hook_extra_parameters:
        # No additional arguments found
        return True, None

    if not allowed_parameters:
        msg = (
            f"'{hook_module}.{hook_name}' has additional attributes "
            f"{', '.join(map(repr, hook_extra_parameters))}. "
            "Expected that all `boto3` related hooks (based on `AwsGenericHook` or `AwsBaseHook`) "
            "should not use additional attributes in class constructor, "
            "please move them to method signatures. "
            f"Make sure that {hook_name!r} constructor has signature `def __init__(self, *args, **kwargs):`"
        )
    else:
        msg = (
            f"'{hook_module}.{hook_name}' allowed only "
            f"{', '.join(map(repr, allowed_parameters))} additional attributes, "
            f"but got extra parameters {', '.join(map(repr, hook_extra_parameters))}. "
            "Please move additional attributes from class constructor into method signatures. "
        )

    return False, msg


@pytest.mark.parametrize("hook_module", get_aws_hooks_modules())
def test_expected_thin_hooks(hook_module: str):
    """
    Test Amazon provider Hooks' signatures.

    All hooks should provide thin wrapper around boto3 / aiobotocore,
    that mean we should not define additional parameters in Hook parameters.
    It should be defined in appropriate methods.

    .. code-block:: python

        # Bad: Thick wrapper
        from airflow.providers.amazon.aws.hooks.base_aws import AwsBaseHook


        class AwsServiceName(AwsBaseHook):
            def __init__(self, foo: str, spam: str, *args, **kwargs) -> None:
                kwargs.update(dict(client_type="service", resource_type=None))
                super().__init__(*args, **kwargs)
                self.foo = foo
                self.spam = spam

            def method1(self):
                if self.foo == "bar":
                    ...

            def method2(self):
                if self.spam == "egg":
                    ...

    .. code-block:: python

        # Good: Thin wrapper
        class AwsServiceName(AwsBaseHook):
            def __init__(self, *args, **kwargs) -> None:
                kwargs.update(dict(client_type="service", resource_type=None))
                super().__init__(*args, **kwargs)

            def method1(self, foo: str):
                if foo == "bar":
                    ...

            def method2(self, spam: str):
                if spam == "egg":
                    ...

    """
    hooks = get_aws_hooks_from_module(hook_module)
    if not hooks:
        pytest.skip(reason=f"Module {hook_module!r} doesn't contain subclasses of `AwsGenericHook`.")

    errors = [
        message
        for valid, message in (validate_hook(hook, hook_name, hook_module) for hook, hook_name in hooks)
        if not valid and message
    ]

    if errors:
        errors_msg = "\n * ".join(errors)
        pytest.fail(reason=f"Found errors in {hook_module}:\n * {errors_msg}")



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import logging
import os
import pathlib
import shutil
from functools import cached_property
from typing import TYPE_CHECKING

import attrs

from airflow.configuration import conf
from airflow.providers.amazon.aws.hooks.s3 import S3Hook
from airflow.providers.amazon.version_compat import AIRFLOW_V_3_0_PLUS
from airflow.utils.log.file_task_handler import FileTaskHandler
from airflow.utils.log.logging_mixin import LoggingMixin

if TYPE_CHECKING:
    from airflow.models.taskinstance import TaskInstance
    from airflow.sdk.types import RuntimeTaskInstanceProtocol as RuntimeTI
    from airflow.utils.log.file_task_handler import LogMessages, LogSourceInfo


@attrs.define
class S3RemoteLogIO(LoggingMixin):  # noqa: D101
    remote_base: str
    base_log_folder: pathlib.Path = attrs.field(converter=pathlib.Path)
    delete_local_copy: bool

    processors = ()

    def upload(self, path: os.PathLike | str, ti: RuntimeTI):
        """Upload the given log path to the remote storage."""
        path = pathlib.Path(path)
        if path.is_absolute():
            local_loc = path
            remote_loc = os.path.join(self.remote_base, path.relative_to(self.base_log_folder))
        else:
            local_loc = self.base_log_folder.joinpath(path)
            remote_loc = os.path.join(self.remote_base, path)

        if local_loc.is_file():
            # read log and remove old logs to get just the latest additions
            log = local_loc.read_text()
            has_uploaded = self.write(log, remote_loc)
            if has_uploaded and self.delete_local_copy:
                shutil.rmtree(os.path.dirname(local_loc))

    @cached_property
    def hook(self):
        """Returns S3Hook."""
        return S3Hook(
            aws_conn_id=conf.get("logging", "REMOTE_LOG_CONN_ID"),
            transfer_config_args={"use_threads": False},
        )

    def s3_log_exists(self, remote_log_location: str) -> bool:
        """
        Check if remote_log_location exists in remote storage.

        :param remote_log_location: log's location in remote storage
        :return: True if location exists else False
        """
        return self.hook.check_for_key(remote_log_location)

    def s3_read(self, remote_log_location: str, return_error: bool = False) -> str:
        """
        Return the log found at the remote_log_location or '' if no logs are found or there is an error.

        :param remote_log_location: the log's location in remote storage
        :param return_error: if True, returns a string error message if an
            error occurs. Otherwise returns '' when an error occurs.
        :return: the log found at the remote_log_location
        """
        try:
            return self.hook.read_key(remote_log_location)
        except Exception as error:
            msg = f"Could not read logs from {remote_log_location} with error: {error}"
            self.log.exception(msg)
            # return error if needed
            if return_error:
                return msg
        return ""

    def write(
        self,
        log: str,
        remote_log_location: str,
        append: bool = True,
        max_retry: int = 1,
    ) -> bool:
        """
        Write the log to the remote_log_location; return `True` or fails silently and return `False`.

        :param log: the contents to write to the remote_log_location
        :param remote_log_location: the log's location in remote storage
        :param append: if False, any existing log file is overwritten. If True,
            the new log is appended to any existing logs.
        :param max_retry: Maximum number of times to retry on upload failure
        :return: whether the log is successfully written to remote location or not.
        """
        try:
            if append and self.s3_log_exists(remote_log_location):
                old_log = self.s3_read(remote_log_location)
                log = f"{old_log}\n{log}" if old_log else log
        except Exception:
            self.log.exception("Could not verify previous log to append")
            return False

        # Default to a single retry attempt because s3 upload failures are
        # rare but occasionally occur.  Multiple retry attempts are unlikely
        # to help as they usually indicate non-ephemeral errors.
        for try_num in range(1 + max_retry):
            try:
                self.hook.load_string(
                    log,
                    key=remote_log_location,
                    replace=True,
                    encrypt=conf.getboolean("logging", "ENCRYPT_S3_LOGS"),
                )
                break
            except Exception:
                if try_num < max_retry:
                    self.log.warning(
                        "Failed attempt to write logs to %s, will retry",
                        remote_log_location,
                    )
                else:
                    self.log.exception("Could not write logs to %s", remote_log_location)
                    return False
        return True

    def read(self, relative_path: str, ti: RuntimeTI) -> tuple[LogSourceInfo, LogMessages | None]:
        logs: list[str] = []
        messages = []
        bucket, prefix = self.hook.parse_s3_url(s3url=os.path.join(self.remote_base, relative_path))
        keys = self.hook.list_keys(bucket_name=bucket, prefix=prefix)
        if keys:
            keys = sorted(f"s3://{bucket}/{key}" for key in keys)
            if AIRFLOW_V_3_0_PLUS:
                messages = keys
            else:
                messages.append("Found logs in s3:")
                messages.extend(f"  * {key}" for key in keys)
            for key in keys:
                logs.append(self.s3_read(key, return_error=True))
            return messages, logs
        return messages, None


class S3TaskHandler(FileTaskHandler, LoggingMixin):
    """
    S3TaskHandler is a python log handler that handles and reads task instance logs.

    It extends airflow FileTaskHandler and uploads to and reads from S3 remote storage.
    """

    def __init__(self, base_log_folder: str, s3_log_folder: str, **kwargs):
        super().__init__(base_log_folder)
        self.handler: logging.FileHandler | None = None
        self.remote_base = s3_log_folder
        self.log_relative_path = ""
        self._hook = None
        self.closed = False
        self.upload_on_close = True
        self.io = S3RemoteLogIO(
            remote_base=s3_log_folder,
            base_log_folder=base_log_folder,
            delete_local_copy=kwargs.get(
                "delete_local_copy", conf.getboolean("logging", "delete_local_logs")
            ),
        )

    def set_context(self, ti: TaskInstance, *, identifier: str | None = None) -> None:
        super().set_context(ti, identifier=identifier)
        # Local location and remote location is needed to open and
        # upload local log file to S3 remote storage.
        if TYPE_CHECKING:
            assert self.handler is not None

        self.ti = ti

        full_path = self.handler.baseFilename
        self.log_relative_path = pathlib.Path(full_path).relative_to(self.local_base).as_posix()
        is_trigger_log_context = getattr(ti, "is_trigger_log_context", False)
        self.upload_on_close = is_trigger_log_context or not getattr(ti, "raw", None)
        # Clear the file first so that duplicate data is not uploaded
        # when reusing the same path (e.g. with rescheduled sensors)
        if self.upload_on_close:
            with open(self.handler.baseFilename, "w"):
                pass

    def close(self):
        """Close and upload local log file to remote storage S3."""
        # When application exit, system shuts down all handlers by
        # calling close method. Here we check if logger is already
        # closed to prevent uploading the log to remote storage multiple
        # times when `logging.shutdown` is called.
        if self.closed:
            return

        super().close()

        if not self.upload_on_close:
            return

        if hasattr(self, "ti"):
            self.io.upload(self.log_relative_path, self.ti)

        # Mark closed so we don't double write if close is called twice
        self.closed = True

    def _read_remote_logs(self, ti, try_number, metadata=None) -> tuple[LogSourceInfo, LogMessages]:
        # Explicitly getting log relative path is necessary as the given
        # task instance might be different than task instance passed in
        # in set_context method.
        worker_log_rel_path = self._render_filename(ti, try_number)

        messages, logs = self.io.read(worker_log_rel_path, ti)

        if logs is None:
            logs = []
            if not AIRFLOW_V_3_0_PLUS:
                messages.append(f"No logs found on s3 for ti={ti}")

        return messages, logs
